from typing import Dict
from typing import List
from typing import Protocol
from typing import TypedDict
from typing import Union

import numpy as np
import torch
from torch.nn.functional import softmax
from torchvision.models.detection import RetinaNet
from torchvision.models.detection.generalized_rcnn import GeneralizedRCNN
from torchvision.models.detection.transform import resize_boxes


class BoxesType(TypedDict):
    boxes: torch.Tensor
    """shape (n, 4); dtype float32.
    Bounding boxes in (x1, y1, x2, y2) format.,
    0 <= x1 < x2 and 0 <= y1 < y2.
    """
    labels: torch.Tensor
    """shape (n,); dtype int64.
    Labels for the detected objects.
    """


class DetectionType(BoxesType):
    scores: torch.Tensor
    """shape (n,); dtype float32.
    Detected class probability for the detected objects.
    """


class KClassDetectionType(DetectionType):
    logits: torch.Tensor
    """shape (n, c); dtype float32.
    K Class logits for the detected objects.
    """


class MasksType(TypedDict):
    tp: np.ndarray
    fp: np.ndarray
    fn: np.ndarray


class TorchDetectorType(Protocol):
    """
    The TorchDetectorType will output a List[DetectionType] and take a List[torch.Tensor] because it is batched.
    """

    def __call__(self, __x: List[torch.Tensor]) -> List[DetectionType]:
        ...


class WrappedTorchDetectorType(Protocol):
    """
    The TorchDetectorType will output a List[DetectionType] and take a List[torch.Tensor] because it is batched.
    """

    def __call__(self, __x: List[torch.Tensor]) -> List[KClassDetectionType]:
        ...


def wrap_detector_with_logit_call(detector: GeneralizedRCNN) -> WrappedTorchDetectorType:
    detector_forward = detector.forward

    def dummy_call(inputs: List[torch.Tensor]) -> List[KClassDetectionType]:
        result = detector_forward(inputs)
        for idx, input in enumerate(inputs):
            input.requires_grad_()
            original_image_size = input.size()[-2:]
            transformed_inputs = detector.transform([input])[0]
            rescaled_image_size = transformed_inputs.image_sizes[0]
            with torch.no_grad():
                rescaled_bbs = resize_boxes(result[idx]["boxes"], original_image_size, rescaled_image_size)
            features = detector.backbone(transformed_inputs.tensors)
            box_features = detector.roi_heads.box_roi_pool(features, [rescaled_bbs], transformed_inputs.image_sizes)
            box_features = detector.roi_heads.box_head(box_features)
            (
                class_logits,
                box_regression,
            ) = detector.roi_heads.box_predictor(box_features)
            result[idx]["logits"] = class_logits
            # probabilities = softmax(class_logits, -1)
            # assert (probabilities[:, 1:].max(-1)[0] == result[idx]["scores"]).all()
            # assert (probabilities[:, 1:].max(-1)[1] == result[idx]["labels"] - 1).all()
            # Why won't these pass??
        return result

    detector.forward = dummy_call
    return detector



import json
from pathlib import Path
from typing import Dict
from typing import List
from typing import Optional
from typing import TypedDict
from typing import Union
from typing import overload

import torch
from torchvision.ops.boxes import box_convert

from hardest.detectors import DetectionType
from hardest.detectors import KClassDetectionType


class AnnotationType(TypedDict):
    """Annotation as a pycocotools target.

    See https://github.com/cocodataset/cocoapi for more information

    """

    bbox: List[float]
    """2d bounding box in (x1, y1, x2, y2) format."""
    iscrowd: bool
    """Indicates whether the box represents a single object or a crowd."""
    category_id: int
    """Labels for the bounding boxes"""


class AnnotationTypeTorch(TypedDict):
    boxes: torch.Tensor
    """shape (n, 4); dtype float32.
    Bounding boxes in (x1, y1, x2, y2) format.,
    0 <= x1 < x2 and 0 <= y1 < y2.
    """
    labels: torch.Tensor
    """shape (n,); dtype int64.
    Labels for the detected objects.
    """
    iscrowd: torch.Tensor
    """shape (n,); dtype bool.
    Class probabilities for the detected objects.
    """


def coco_api_to_torchvision(coco_ann: List[AnnotationType]) -> AnnotationTypeTorch:
    """
    Convert python coco api targets to torchvision targets
    Args:
        coco_ann: Annotations in pycocotools format

    Returns:
        Annotations in torchvision format
    """
    boxes_coco = torch.Tensor([ann["bbox"] for ann in coco_ann])  # x, y, w, h
    boxes_pytorch = box_convert(boxes_coco, "xywh", "xyxy") if len(boxes_coco) > 0 else boxes_coco
    iscrowd = torch.Tensor([ann["iscrowd"] for ann in coco_ann])
    return {
        "boxes": boxes_pytorch,
        "labels": torch.Tensor([ann["category_id"] for ann in coco_ann]),
        "iscrowd": iscrowd,
    }


@overload
def filter_pytorch_detections_by_score(detections: KClassDetectionType, score_threshold: float) -> KClassDetectionType:
    ...


@overload
def filter_pytorch_detections_by_score(detections: DetectionType, score_threshold: float) -> DetectionType:
    ...


def filter_pytorch_detections_by_score(
    detections: Union[DetectionType, KClassDetectionType], score_threshold: float
) -> Union[DetectionType, KClassDetectionType]:
    """
    Applies a score threshold to torchvision detections
    Args:
        detections: Detections to be filtered in torchvision format
        score_threshold: Only detections with scores above the score threshold will be kept

    Returns:
        filtered detections
    """
    keep_mask = detections["scores"] >= score_threshold
    return {key: value[keep_mask] for key, value in detections.items()}  # type: ignore
    # ignore required until https://github.com/python/mypy/issues/7981 is fixed


def serialise_detections(save_path: Path, detections: List[DetectionType]):
    """
    Saves torchvision detections into a json
    Args:
        save_path: path to json where detections should be saved
        detections: detections to be saved

    """
    exportable_detections = [{key: value.cpu().numpy().tolist() for key, value in det.items()} for det in detections]  # type: ignore
    # ignore required until https://github.com/python/mypy/issues/7981 is fixed
    with open(save_path, "w") as output_file:
        json.dump(exportable_detections, output_file)


def load_detections(load_path: Path) -> List[DetectionType]:
    """
    Loads torchvision detections from a json
    Args:
        load_path: path to json where detections should be loaded from

    Returns:
        loaded detections

    """
    with open(load_path, "r") as load_file:
        exportable_detections = json.load(load_file)
    return [{key: torch.Tensor(value) for key, value in det.items()} for det in exportable_detections]  # type: ignore
    # ignore required until https://github.com/python/mypy/issues/7981 is fixed


NUSCENES_CATEGORIES_MMDET = (
    "car",
    "truck",
    "trailer",
    "bus",
    "construction_vehicle",
    "bicycle",
    "motorcycle",
    "pedestrian",
    "traffic_cone",
    "barrier",
)

TO_AV_SIMPLE = {
    # mmdet
    "car": "Vehicle",
    "truck": "Vehicle",
    "trailer": "Vehicle",
    "bus": "Vehicle",
    "construction_vehicle": "Vehicle",
    "bicycle": "Vehicle",
    "motorcycle": "Vehicle",
    "pedestrian": "Pedestrian",
    "traffic_cone": None,
    "barrier": None,
}

CONTIGUOUS_ID_OF_AVSIMPLE_NAME = {
    "Vehicle": 0,
    "Pedestrian": 1,
}


def nuimages_coco_mmdet_to_av_simple(annotations: Optional[AnnotationTypeTorch]) -> Optional[AnnotationTypeTorch]:
    """
    Converts labels for annotations in mmdet schema to a simplified schema.
    Args:
        annotations: Annotations in the mmdet schema

    Returns:
        Annotations in a simplified schema

    """
    if annotations is None:
        return None
    if len(annotations["labels"]) == 0:
        return annotations
    label = annotations["labels"]
    names = [NUSCENES_CATEGORIES_MMDET[int(instance.item())] for instance in label]
    av_simple_names = [TO_AV_SIMPLE[n] for n in names]
    mask = torch.tensor([n is not None for n in av_simple_names])

    new_labels = torch.tensor([-1 if n is None else CONTIGUOUS_ID_OF_AVSIMPLE_NAME[n] for n in av_simple_names])
    annotations["labels"] = new_labels
    for key, value in annotations.items():
        annotations[key] = value[mask]  # type: ignore
    # ignore required until https://github.com/python/mypy/issues/7981 is fixed
    return annotations



from abc import ABC
from abc import abstractmethod
from typing import List
from typing import Tuple

import numpy as np
import torch
from torch.distributions import Bernoulli

from hardest.hardness_definitions import DetectionType
from hardest.utils import AnnotationTypeTorch
from hardest.utils import filter_pytorch_detections_by_score


class GroundTruthSampler(ABC):
    """
    Abstract class for sampling based pseudo ground truth generation.
    """

    @abstractmethod
    def sample(self, detections: DetectionType, n_samples: int) -> List[AnnotationTypeTorch]:
        """
        Sample the pseudo ground truth.
        Args:
            detections: detections for a single image
            n_samples: number of samples to generate

        Returns:
            Sampled pseudo ground truth in torchvision format
        """
        pass


class ScoreSampling(GroundTruthSampler):
    """
    Score sampling pseudo ground truth generation.

    Produces samples of pseudo ground truth using detections which are included based on a Bernouilli sampled mask from
    the detection scores.
    """

    def sample(self, detections: DetectionType, n_samples: int) -> List[AnnotationTypeTorch]:
        """
        Sample the pseudo ground truth.
        Args:
            detections: detections for a single image
            n_samples: number of samples to generate

        Returns:
            Sampled pseudo ground truth in torchvision format
        """
        keep_distribution = Bernoulli(probs=detections["scores"])
        keep_mask = keep_distribution.sample((n_samples,)).bool()
        psuedo_ground_truth: List[AnnotationTypeTorch] = [
            {
                "boxes": detections["boxes"][keep_sample],
                "labels": detections["labels"][keep_sample],
                "iscrowd": torch.zeros_like(detections["labels"][keep_sample], dtype=torch.bool),
            }
            for keep_sample in keep_mask
        ]
        return psuedo_ground_truth


class DiscreteCommittee(ABC):
    """
    A discrete distribution for pseudo ground truth using detections from several detectors
    """

    @abstractmethod
    def probability_mass(self, detection_list: List[DetectionType]) -> Tuple[List[AnnotationTypeTorch], List[float]]:
        pass


class UniformCommittee(DiscreteCommittee):
    """
    A uniform distribution for pseudo ground truth using detections from several detectors

    Args:
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
    """

    def __init__(self, score_threshold: float = 0.5):
        self.score_threshold: float = score_threshold

    def probability_mass(self, detection_list: List[DetectionType]) -> Tuple[List[AnnotationTypeTorch], List[float]]:
        """
        Probability mass of detections in discrete distribution.
        Args:
            detection_list: List of detections for a single image, list over committee members.

        Returns:
            Pseudo ground truth detections in torchvision format
            List of probability mass for each detection
        """
        detection_list = [filter_pytorch_detections_by_score(det, self.score_threshold) for det in detection_list]
        psuedo_ground_truth: List[AnnotationTypeTorch] = [
            {
                "boxes": detections["boxes"],
                "labels": detections["labels"],
                "iscrowd": torch.zeros_like(detections["labels"], dtype=torch.bool),
            }
            for detections in detection_list
        ]
        return psuedo_ground_truth, len(detection_list) * [1 / len(detection_list)]


class ScoreSamplingUniformCommittee(ABC):
    """
    Uniform committee with score sampling for the committee members.
    """

    def sample(self, detection_list: List[DetectionType], n_samples: int) -> List[AnnotationTypeTorch]:
        """
        Sample the pseudo ground truth.
        Args:
            detection_list: List of detections for a single image, list over committee members.
            n_samples: number of samples to generate

        Returns:
            Sampled pseudo ground truth in torchvision format
        """
        detections, probability_mass = self.underlying_committee_model(detection_list)
        sample_mask = np.random.choice(len(detections), size=n_samples, p=probability_mass)
        sampled_detections = [detections[idx] for idx in sample_mask]
        return [ScoreSampling().sample(det, n_samples=n_samples)[0] for det in sampled_detections]

    def underlying_committee_model(
        self, detection_list: List[DetectionType]
    ) -> Tuple[List[DetectionType], List[float]]:
        """
        Probability mass of detections in discrete distribution for underlying committe.

        Note that this is NOT the distribution for this pseudo ground truth model, since a continious sampling step is
        applied to the results of this method.
        Args:
            detection_list: List of detections for a single image, list over committee members.

        Returns:
            Pseudo ground truth detections in torchvision format
            List of probability mass for each detection
        """
        return detection_list, (np.ones(len(detection_list)) / len(detection_list)).tolist()



import contextlib
import copy
from collections import defaultdict
from contextlib import redirect_stdout
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TypedDict
from typing import Union

import numpy as np
import torch
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval
from torchvision.ops import box_area
from torchvision.ops import box_convert
from torchvision.ops.boxes import _box_inter_union

from hardest.detectors import BoxesType
from hardest.detectors import MasksType
from hardest.hardness_definitions import AnnotationTypeTorch
from hardest.hardness_definitions import DetectionType
from hardest.hardness_definitions import HardnessDefinition

COCO_CAT_IDS = set(range(91))
AV_SIMPLE_CAT_IDS = set(range(2))

# Note that this file contains lots of copied boilerplate code from
# https://github.com/pytorch/vision/tree/main/references/detection which has never been properly tested or integrated
# into torchvision, in addition to copied code from pycocotools


def no_weighting(instances: BoxesType, masks: MasksType, image: torch.Tensor) -> torch.Tensor:
    """
    Equal weight for each detection
    Args:
        instances: Box instance for a single image.
        masks: TP/FP/FN masks for the box instances.
        image: Image for which hardness is being computed in torchvision format (C, H, W)

    Returns:
        weighting
    """
    return torch.ones_like(instances["labels"])


def pixel(instances: BoxesType, masks: MasksType, image: torch.Tensor) -> torch.Tensor:
    """
    Weight for each detection based on area relative to whole image area
    Args:
        instances: Box instance for a single image.
        masks: TP/FP/FN masks for the box instances.
        image: Image for which hardness is being computed in torchvision format (C, H, W)

    Returns:
        weighting
    """
    if len(instances["labels"]) > 0:
        image_area = image.shape[-1] * image.shape[-2]
        weights_det = box_area(instances["boxes"]) / image_area
    else:
        weights_det = torch.ones_like(instances["labels"])

    return weights_det


def overlap(instances: BoxesType, masks: MasksType, image: torch.Tensor) -> torch.Tensor:
    """
    Weight for each detection equal to number of overlapping TP boxes (can be fractional)
    Args:
        instances: Box instance for a single image.
        masks: TP/FP/FN masks for the box instances.
        image: Image for which hardness is being computed in torchvision format (C, H, W)

    Returns:
        weighting
    """
    if len(instances["labels"]) > 0:
        assert instances["boxes"].shape[0] == masks["tp"].shape[0]
        intersection_all_det, _ = _box_inter_union(
            instances["boxes"], instances["boxes"][torch.Tensor(masks["tp"]).bool()]
        )
        weights_det = (
            intersection_all_det.sum(axis=1) / box_area(instances["boxes"]) - masks["tp"]
        )  # subtract one to avoid counting intersection with itself when computing weights for TPs
    else:
        weights_det = torch.ones_like(instances["labels"])

    return weights_det


weightings: Dict[Optional[str], Callable[[BoxesType, MasksType, torch.Tensor], torch.Tensor]] = {
    None: no_weighting,
    "overlap": overlap,
    "pixel": pixel,
}


class CocoHardness(HardnessDefinition):
    """
    Parent class for hardness definitions based on the coco evaluation

    Args:
        cat_ids: Category ids for labels. Will default to coco ids if this is not set, which could cause errors.
        area_rng: Range of bounding box areas to be considered for evaluation (default is maximum range in coco).
        max_dets: Number of detections to be used in evaluation (extra detections will be trimmed).
        iou_threshold: Only boxes with an intersection over union larger than iou_threshold can be associated.
        weighting_fn: Weighting function to be used for boxes in evaluation when computing hardness.
            See hardest.pycocotools.weightings

    Attributes:
        cat_ids: Category ids for labels. Will default to coco ids if this is not set, which could cause errors.
        area_rng: Range of bounding box areas to be considered for evaluation (default is maximum range in coco).
        max_dets: Number of detections to be used in evaluation (extra detections will be trimmed).
        iou_threshold: Only boxes with an intersection over union larger than iou_threshold can be associated.
        weighting_fn: Weighting function to be used for boxes in evaluation when computing hardness.
            See hardest.pycocotools.weightings
    """

    def __init__(
        self,
        cat_ids: Set[int] = COCO_CAT_IDS,
        area_rng: Tuple[float, float] = (0.0, 10000000000.0),
        max_dets: int = 100,
        iou_threshold: float = 0.5,
        weighting_fn: Callable[[BoxesType, MasksType, torch.Tensor], torch.Tensor] = no_weighting,
    ):
        self.cat_ids: Set[int] = cat_ids
        self.area_rng: List[float] = list(area_rng)
        self.max_dets: int = max_dets
        self.iou_threshold: float = iou_threshold
        self.weighting_fn: Callable[[BoxesType, MasksType, torch.Tensor], torch.Tensor] = weighting_fn

    def __call__(self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor) -> float:
        """
        Abstract method to compute hardness for image
        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness for image
        """
        raise NotImplementedError

    def get_masks(
        self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        False positive and false negative mask for single image

        Largely copied from:
        https://github.com/pytorch/vision/blob/de31e4b8bf9b4a7e0668d19059a5ac4760dceee1/references/detection/coco_eval.py#L28

        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            False positive binary mask
            False negative binary mask
        """
        with redirect_stdout(None):
            coco_gt = convert_to_coco_api(copy.deepcopy(annotations), image, categories=self.cat_ids)

            coco_eval = COCOeval(coco_gt, iouType="bbox")
            coco_eval.params.maxDets = [self.max_dets]
            coco_eval.params.iouThrs = [self.iou_threshold]

            results = prepare_for_coco_detection(detections)
            coco_dt = loadRes(coco_gt, results) if results else COCO()

            coco_eval.cocoDt = coco_dt
            coco_eval.params.imgIds = [0]
            coco_eval.params.areaRng = [
                self.area_rng,
            ]
            coco_eval.evaluate()
            eval_imgs = coco_eval.evalImgs
            fp_mask, fn_mask = accumulate(
                eval_imgs, list(self.cat_ids), n_gt=len(annotations["labels"]), n_det=len(detections["labels"])
            )
            return fp_mask, fn_mask

    def get_weights(
        self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor
    ) -> Tuple[MasksType, torch.Tensor]:
        """
        Compute TP/FP/FN masks and associated weights for all box instances in an image
        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            TP/FP/FN masks.
            Associated weights.
        """
        fp_mask, fn_mask = self.get_masks(detections, annotations, image)

        nfn = np.sum(fn_mask)

        instances: BoxesType = {
            "boxes": torch.cat((detections["boxes"], annotations["boxes"][torch.Tensor(fn_mask).bool()])),
            "labels": torch.cat((detections["labels"], annotations["labels"][torch.Tensor(fn_mask).bool()])),
        }

        masks: MasksType = {
            "tp": np.concatenate((~fp_mask, np.zeros(nfn, dtype=bool))),
            "fp": np.concatenate((fp_mask, np.zeros(nfn, dtype=bool))),
            "fn": np.concatenate((np.zeros_like(fp_mask), np.ones(nfn, dtype=bool))),
        }

        weights = self.weighting_fn(instances, masks, image)

        assert (weights >= 0).all()

        return masks, weights


class FP(CocoHardness):
    """
    Pycocotools false positives (weighted)
    """

    def __call__(self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor) -> float:
        """
        Computes weighted false positive score

        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness for image
        """
        masks, weights = self.get_weights(detections, annotations, image)
        return np.sum(masks["fp"] * weights.numpy()).item()


class FN(CocoHardness):
    """
    Pycocotools coco false negatives (weighted)
    """

    def __call__(self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor) -> float:
        """
        Computes weighted false negative score

        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness for image
        """
        masks, weights = self.get_weights(detections, annotations, image)
        return np.sum(masks["fn"] * weights.numpy()).item()


class TotalFalse(CocoHardness):
    """
    Pycocotools coco total false (weighted)
    """

    def __call__(self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor) -> float:
        """
        Computes weighted total false score

        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness for image
        """
        masks, weights = self.get_weights(detections, annotations, image)
        return np.sum(np.logical_or(masks["fn"], masks["fp"]) * weights.numpy()).item()


# The below functions are copied with minor modification from pycocotools and torchvision tutorials


def accumulate(
    eval_images: List[Dict[str, Any]], cat_ids: List[int], n_gt: int, n_det: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert eval images for every category into a mask for detections and annotations

    Copied from https://github.com/cocodataset/cocoapi/blob/6c3b394c07aed33fd83784a8bf8798059a1e9ae4/PythonAPI/pycocotools/cocoeval.py#L315
    Modified to remove code unrelated to producing mask and to produce fn

    Args:
        eval_images: Evaluation results from pycocotools
        cat_ids: List of category (label) ids
        n_gt: Number of annotations
        n_det: Number of detections

    Returns:
        False positive binary mask
        False negative binary mask

    """
    fp_mask = np.zeros((n_det,), dtype=bool)
    fn_mask = np.zeros((n_gt,), dtype=bool)

    for k, _ in enumerate(cat_ids):
        if eval_images[k] is None:
            continue

        dtm = eval_images[k]["dtMatches"]
        gtm = eval_images[k]["gtMatches"]
        dtIg = eval_images[k]["dtIgnore"]
        gtIg = eval_images[k]["gtIgnore"]
        if len(eval_images[k]["dtIds"]) > 0:
            fps = np.logical_and(np.logical_not(dtm), np.logical_not(dtIg))
            fp_mask[np.array(eval_images[k]["dtIds"]) - 1] = fps
        if len(eval_images[k]["gtIds"]) > 0:
            fns = np.logical_and(np.logical_not(gtm), np.logical_not(gtIg))
            fn_mask[np.array(eval_images[k]["gtIds"]) - 1] = fns
    return fp_mask, fn_mask


class CocoDetection(TypedDict):
    image_id: int
    category_id: float
    bbox: List[float]
    score: float
    area: Optional[float]
    id: Optional[int]
    iscrowd: Optional[int]


def prepare_for_coco_detection(predictions: DetectionType) -> List[CocoDetection]:
    """
    Convert torchvision detections into pycocotools evaluation format

    Bounding boxes are converted from xyxy to xywh. Other attributes are converted to list from tensor and renamed.
    Copied from: https://github.com/pytorch/vision/blob/de31e4b8bf9b4a7e0668d19059a5ac4760dceee1/references/detection/coco_eval.py#L67

    Args:
        predictions: Detections for a single image in torchvision format

    Returns:
        pycocotools format detections
    """
    labels = predictions["labels"].tolist()
    n_det = len(labels)
    boxes = predictions["boxes"]
    if n_det > 0:
        boxes = box_convert(boxes, "xyxy", "xywh").tolist()
    scores = predictions["scores"].tolist()

    return [
        {
            "image_id": 0,
            "category_id": labels[k],
            "bbox": box,
            "score": scores[k],
            "iscrowd": None,
            "id": None,
            "area": None,
        }
        for k, box in enumerate(boxes)
    ]


def loadRes(coco_gt: COCO, anns: List[CocoDetection]) -> COCO:
    """
    Load result detections and return a result api object.

    Modified from:
    https://github.com/cocodataset/cocoapi/blob/6c3b394c07aed33fd83784a8bf8798059a1e9ae4/PythonAPI/pycocotools/coco.py#L297
    to avoid loading res from hard disk
    Args:
        coco_gt: annotation api object
        anns: the list of prediction results

    Returns:
        result api object
    """
    res = COCO()
    res.dataset["images"] = [img for img in coco_gt.dataset["images"]]

    assert type(anns) == list, "results in not an array of objects"
    annsImgIds = [ann["image_id"] for ann in anns]
    assert set(annsImgIds) == (
        set(annsImgIds) & set(coco_gt.getImgIds())
    ), "Results do not correspond to current coco set"

    if "bbox" in anns[0] and not anns[0]["bbox"] == []:
        res.dataset["categories"] = copy.deepcopy(coco_gt.dataset["categories"])
        for id, ann in enumerate(anns):
            bb = ann["bbox"]
            ann["area"] = bb[2] * bb[3]
            ann["id"] = id + 1
            ann["iscrowd"] = 0
    else:
        raise ValueError("bbox must be present in annotations")

    res.dataset["annotations"] = anns
    res.createIndex()
    return res


def convert_to_coco_api(target: AnnotationTypeTorch, img: torch.Tensor, categories: Set[int]) -> COCO:
    """
    Convert annotatinos to coco api

    Copied from: https://github.com/pytorch/vision/blob/de31e4b8bf9b4a7e0668d19059a5ac4760dceee1/references/detection/coco_utils.py#L146
    with modification to remove unused segmentation and keypoints and to work on annotations
    Args:
        target: Annotations for a single image in torchvision format
        img: Image for which hardness is being computed in torchvision format (C, H, W)
        categories: List of category (label) ids

    Returns:
        result api object

    """
    coco_ds = COCO()
    # annotation IDs need to start at 1, not 0, see torchvision issue #1530
    ann_id = 1
    dataset: Dict[str, List[Dict[str, Any]]] = {"images": [], "categories": [], "annotations": []}
    image_id = 0
    img_dict = {}
    img_dict["id"] = image_id
    img_dict["height"] = img.shape[-2]
    img_dict["width"] = img.shape[-1]
    dataset["images"].append(img_dict)
    num_objs = len(target["boxes"])
    if num_objs > 0:
        bboxes = box_convert(target["boxes"], "xyxy", "xywh").tolist()
        labels = target["labels"].tolist()
        areas = box_area(target["boxes"]).tolist()
        iscrowd = target["iscrowd"].tolist()
    for i in range(num_objs):
        assert labels[i] in categories
        ann = {
            "image_id": image_id,
            "bbox": bboxes[i],
            "category_id": labels[i],
            "area": areas[i],
            "iscrowd": iscrowd[i],
            "id": ann_id,
        }
        dataset["annotations"].append(ann)
        ann_id += 1
    dataset["categories"] = [{"id": i} for i in sorted(categories)]
    coco_ds.dataset = dataset
    with contextlib.redirect_stdout(None):
        coco_ds.createIndex()
    return coco_ds


hardness_definition_factory: Dict[str, HardnessDefinition] = {
    "fp": FP(),
    "fn": FN(),
    "false": TotalFalse(),
    "fp_pixel": FP(weighting_fn=pixel),
    "fn_pixel": FN(weighting_fn=pixel),
    "false_pixel": TotalFalse(weighting_fn=pixel),
    "fp_overlap": FP(weighting_fn=overlap),
    "fn_overlap": FN(weighting_fn=overlap),
    "false_overlap": TotalFalse(weighting_fn=overlap),
}

hardness_definition_factory_av_simple: Dict[str, HardnessDefinition] = {
    "fp": FP(cat_ids=AV_SIMPLE_CAT_IDS),
    "fn": FN(cat_ids=AV_SIMPLE_CAT_IDS),
    "false": TotalFalse(cat_ids=AV_SIMPLE_CAT_IDS),
    "fp_pixel": FP(weighting_fn=pixel, cat_ids=AV_SIMPLE_CAT_IDS),
    "fn_pixel": FN(weighting_fn=pixel, cat_ids=AV_SIMPLE_CAT_IDS),
    "false_pixel": TotalFalse(weighting_fn=pixel, cat_ids=AV_SIMPLE_CAT_IDS),
    "fp_overlap": FP(weighting_fn=overlap, cat_ids=AV_SIMPLE_CAT_IDS),
    "fn_overlap": FN(weighting_fn=overlap, cat_ids=AV_SIMPLE_CAT_IDS),
    "false_overlap": TotalFalse(weighting_fn=overlap, cat_ids=AV_SIMPLE_CAT_IDS),
}



from abc import ABC
from abc import abstractmethod

import torch

from hardest.detectors import DetectionType
from hardest.utils import AnnotationTypeTorch


class HardnessDefinition(ABC):
    """
    Abstract class for hardness definitions
    """

    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, detections: DetectionType, annotations: AnnotationTypeTorch, image: torch.Tensor) -> float:
        """
        Calculate hardness of image
        Args:
            detections: Detections for a single image in torchvision format
            annotations: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness

        """
        raise NotImplementedError



import itertools
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Iterable
from typing import List
from typing import Literal
from typing import Union

import numpy as np
import torch
from scipy.special import entr
from scipy.stats import entropy
from torch.distributions import Bernoulli
from torch.nn.functional import softmax
from tqdm import tqdm

from hardest.detectors import BoxesType
from hardest.detectors import DetectionType
from hardest.detectors import KClassDetectionType
from hardest.detectors import MasksType
from hardest.hardness_definitions import HardnessDefinition
from hardest.pseudo_ground_truth_models import DiscreteCommittee
from hardest.pseudo_ground_truth_models import GroundTruthSampler
from hardest.pseudo_ground_truth_models import ScoreSampling
from hardest.pseudo_ground_truth_models import ScoreSamplingUniformCommittee
from hardest.pycocotools import no_weighting
from hardest.utils import AnnotationTypeTorch
from hardest.utils import filter_pytorch_detections_by_score


class SupervisedHardnessCalculation:
    """
    Computes hardness according to a specific definition for a 2d object detection dataset based on annotations.

    Args:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Attributes:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
    """

    def __init__(
        self,
        hardness_definition: HardnessDefinition,
        score_threshold: float = 0.5,
    ):
        self.score_threshold: float = score_threshold
        self.hardness_definition: HardnessDefinition = hardness_definition

    def calculate_hardness(self, detection: DetectionType, target: AnnotationTypeTorch, image: torch.Tensor) -> float:
        """
        Calculate the hardness for a single image
        Args:
            detection: Detections for a single image in torchvision format
            target: Annotations for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness as a float
        """
        detection = filter_pytorch_detections_by_score(detection, self.score_threshold)
        return self.hardness_definition(detection, target, image)

    def eval_dataset(
        self,
        detections: Iterable[DetectionType],
        targets: Iterable[AnnotationTypeTorch],
        images: Iterable[torch.Tensor],
    ) -> List[float]:
        """
        Scores an entire dataset by hardness.
        Args:
            detections: List of detections in torchvision format
            targets: List of targets in torchvision format
            images: List of images in torchvision format (C, H, W) (can be a generator if dataset is large)

        Returns:
            Hardness of dataset
        """
        with tqdm(detections, desc=f"Calculating hardness with {self}, fraction of dataset processed") as bar:
            return list(itertools.starmap(self.calculate_hardness, zip(bar, targets, images)))


class UnsupervisedHardnessCalculation(ABC):
    """
    Abstract class for hardness estimation strategies when targets/annotations are not available
    """

    @abstractmethod
    def calculate_hardness(self, detection: DetectionType, image: torch.Tensor) -> float:
        """
        Calculate the hardness for a single image
        Args:
            detection: Detections for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Estimated hardness
        """
        raise NotImplementedError

    def eval_dataset(self, detections: Iterable[DetectionType], images: Iterable[torch.Tensor]) -> List[float]:
        """
        Scores an entire dataset by hardness.
        Args:
            detections: List of detections in torchvision format (should be obtained by running the detector with a
                low score threshold as low score detections are required for the algorithm to work correctly).
            images: List of images in torchvision format (C, H, W) (can be a generator if dataset is large)

        Returns:
            Estimated hardness of dataset
        """
        with tqdm(detections, desc=f"Calculating hardness with {self}, fraction of dataset processed") as bar:
            return list(itertools.starmap(self.calculate_hardness, zip(bar, images)))


class UncertaintyMeasure(ABC):
    """
    Abstract class for uncertainty functions from detected boxes in an image.

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.

    Attributes:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
    """

    def __init__(
        self,
        reduction: Callable[[np.ndarray], np.ndarray] = np.sum,
    ):
        self.reduction: Callable[[np.ndarray], np.ndarray] = reduction

    @abstractmethod
    def __call__(self, detection: DetectionType) -> float:
        pass


class KClassUncertaintyMeasure(ABC):
    """
    Abstract class for uncertainty functions from detected boxes in an image.

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.

    Attributes:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
    """

    def __init__(
        self,
        reduction: Callable[[np.ndarray], np.ndarray] = np.sum,
    ):
        self.reduction: Callable[[np.ndarray], np.ndarray] = reduction

    @abstractmethod
    def __call__(self, detection: KClassDetectionType) -> float:
        pass


class UncertaintyMeasureHardnessCalculation(UnsupervisedHardnessCalculation):
    """
    Unsupervised hardness calculation with a uncertainty measure which is a function of detections.

    Parent class for unsupervised hardness calculations where the calculated hardness does not aim to estimate a
    specific definition of hardness which could be calculated exactly if the targets were available. In this sense they
    are "generic" to any hardness definition.

    Args:
        uncertainty_measure: A function from detections to hardness
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Attributes:
        uncertainty_measure: A function from detections to hardness
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
    """

    def __init__(
        self,
        uncertainty_measure: UncertaintyMeasure,
        score_threshold: float = 0.5,
    ):
        super().__init__()
        self.uncertainty_measure: UncertaintyMeasure = uncertainty_measure
        self.score_threshold: float = score_threshold

    def calculate_hardness(self, detection: DetectionType, image: torch.Tensor) -> float:
        """
        Calculate the hardness for a single image
        Args:
            detections: List of detections in torchvision format (should be obtained by running the detector with a
                low score threshold as low score detections are required for the algorithm to work correctly).
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Estimated hardness
        """
        detection = filter_pytorch_detections_by_score(detection, self.score_threshold)
        return self.uncertainty_measure(detection)


class KClassUncertaintyMeasureHardnessCalculation(ABC):
    """
    Unsupervised hardness calculation with a uncertainty measure which is a function of detections.

    Parent class for unsupervised hardness calculations where the calculated hardness does not aim to estimate a
    specific definition of hardness which could be calculated exactly if the targets were available. In this sense they
    are "generic" to any hardness definition.

    Args:
        uncertainty_measure: A function from detections to hardness
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Attributes:
        uncertainty_measure: A function from detections to hardness
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
    """

    def __init__(
        self,
        uncertainty_measure: KClassUncertaintyMeasure,
        score_threshold: float = 0.5,
    ):
        super().__init__()
        self.uncertainty_measure: KClassUncertaintyMeasure = uncertainty_measure
        self.score_threshold: float = score_threshold

    def calculate_hardness(self, detection: KClassDetectionType, image: torch.Tensor) -> float:
        """
        Calculate the hardness for a single image
        Args:
            detections: List of detections in torchvision format (should be obtained by running the detector with a
                low score threshold as low score detections are required for the algorithm to work correctly).
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Estimated hardness
        """
        detection = filter_pytorch_detections_by_score(detection, self.score_threshold)
        return self.uncertainty_measure(detection)

    def eval_dataset(self, detections: Iterable[KClassDetectionType], images: Iterable[torch.Tensor]) -> List[float]:
        """
        Scores an entire dataset by hardness.
        Args:
            detections: List of detections in torchvision format (should be obtained by running the detector with a
                low score threshold as low score detections are required for the algorithm to work correctly).
            images: List of images in torchvision format (C, H, W) (can be a generator if dataset is large)

        Returns:
            Estimated hardness of dataset
        """
        with tqdm(detections, desc=f"Calculating hardness with {self}, fraction of dataset processed:") as bar:
            return list(itertools.starmap(self.calculate_hardness, zip(bar, images)))


class BinaryDempster(UncertaintyMeasure):
    r"""
    Dempster Shafer (evidential deep learning) uncertainty measure.

    Calculates hardness using evidential deep learning based on a modified version of the method provided in:
    https://arxiv.org/abs/1806.01768

    We modify the method to only consider two classes (detected class vs all other classes).
    Hence unassigned belief is calculated as:
    .. math::
        u = \sum_{b \in B} \frac{|K|}{\sum_{k \in K} 1 + \exp s_{b,k}}
    with  |K| = 2 , where s_{b,k} are the
    class scores for box $b$.
    """

    def __call__(self, detection: DetectionType) -> float:
        r"""
        Calculates uncertainty, i.e. unassigned belief, in the dempster shafer framework.
        .. math::
            u = \sum_{b \in B} \frac{|K|}{\sum_{k \in K} 1 + \exp s_{b,k}}
        with  |K| = 2 , where s_{b,k} are the
        class scores for box $b$.
        Args:
            detection: Detections for a single image in torchvision format

        Returns:
            Dempster score

        """
        p = detection["scores"]
        Z = np.sum([p / (1 - p), (1 - p) / p], axis=0)
        DS = 2 / (2 + Z)
        return self.reduction(DS.numpy()).item()


class Dempster(KClassUncertaintyMeasure):
    r"""
    Dempster Shafer (evidential deep learning) uncertainty measure.

    Calculates hardness using evidential deep learning based on a modified version of the method provided in:
    https://arxiv.org/abs/1806.01768

    We modify the method to only consider two classes (detected class vs all other classes).
    Hence unassigned belief is calculated as:
    .. math::
        u = \sum_{b \in B} \frac{|K|}{\sum_{k \in K} 1 + \exp s_{b,k}}
    with  |K| = 2 , where s_{b,k} are the
    class scores for box $b$.
    """

    def __call__(self, detection: KClassDetectionType) -> float:
        r"""
        Calculates uncertainty, i.e. unassigned belief, in the dempster shafer framework.
        .. math::
            u = \sum_{b \in B} \frac{|K|}{\sum_{k \in K} 1 + \exp s_{b,k}}
        with  |K| = 2 , where s_{b,k} are the
        class scores for box $b$.
        Args:
            detection: Detections for a single image in torchvision format

        Returns:
            Dempster score

        """
        logits = detection["logits"].numpy()
        K = logits.shape[-1]
        Z = np.sum(np.exp(logits), axis=-1)
        DS = K / (K + Z)
        return self.reduction(DS).item() if len(logits) > 0 else float(0.0)


def BinaryDempsterHardnessCalculation(
    reduction: Callable[[np.ndarray], np.ndarray] = np.sum, score_threshold: float = 0.5
) -> UncertaintyMeasureHardnessCalculation:
    """
    Convenience function for hardness calculation with Dempster uncertainty measure

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Returns:
        Uncertainty measure hardness calculation

    """
    return UncertaintyMeasureHardnessCalculation(BinaryDempster(reduction), score_threshold)


def DempsterHardnessCalculation(
    reduction: Callable[[np.ndarray], np.ndarray] = np.sum, score_threshold: float = 0.5
) -> KClassUncertaintyMeasureHardnessCalculation:
    """
    Convenience function for hardness calculation with Dempster uncertainty measure

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Returns:
        Uncertainty measure hardness calculation

    """
    return KClassUncertaintyMeasureHardnessCalculation(Dempster(reduction), score_threshold)


class MaxUncertainty(UncertaintyMeasure):
    """
    Hardness defined as 1-score of the most uncertain bounding box in the image

    Defined in from https://arxiv.org/abs/1801.05124 sec 3.1
    """

    def __call__(self, detection: DetectionType) -> float:
        """
        1-score of the most uncertain bounding box
        Args:
            detection: Detections for a single image in torchvision format

        Returns:
            Most uncertain bounding box uncertainty

        """
        p = detection["scores"]
        uncertainty = 1 - p
        return max(uncertainty).item() if len(p) > 0 else float(0.0)


def MaxUncertaintyCalculation(
    reduction: Callable[[np.ndarray], np.ndarray] = np.sum, score_threshold: float = 0.5
) -> UncertaintyMeasureHardnessCalculation:
    """
    Convenience function for hardness calculation with MaxUncertainty uncertainty measure

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Returns:
        Uncertainty measure hardness calculation

    """
    return UncertaintyMeasureHardnessCalculation(MaxUncertainty(reduction), score_threshold)


class BinaryEntropy(UncertaintyMeasure):
    """
    Hardness defined by the binary entropy for the score for each bounding box
    """

    def __call__(self, detection: DetectionType) -> float:
        r"""
        .. math::
            \sum_{b \in B} \sum_{k \in K} - p_{b,k} \log p_{b,k}
        Args:
            detection: Detections for a single image in torchvision format

        Returns:
            Binary entropy
        """
        p = detection["scores"].numpy()
        return float(self.reduction(entr(p) + entr(1 - p)))


class Entropy(KClassUncertaintyMeasure):
    """
    Hardness defined by the entropy for the k class probabilities for each bounding box
    """

    def __call__(self, detection: KClassDetectionType) -> float:
        r"""
        .. math::
            \sum_{b \in B} \sum_{k \in K} - p_{b,k} \log p_{b,k}
        Args:
            detection: Detections for a single image in torchvision format

        Returns:
            Binary entropy
        """
        p = softmax(detection["logits"], -1).numpy()
        return float(self.reduction(entropy(p, axis=-1))) if len(p) > 0 else float(0.0)


def EntropyCalculation(
    reduction: Callable[[np.ndarray], np.ndarray] = np.sum, score_threshold: float = 0.5
) -> KClassUncertaintyMeasureHardnessCalculation:
    """
    Convenience function for hardness calculation with Entropy uncertainty measure

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Returns:
        Uncertainty measure hardness calculation

    """
    return KClassUncertaintyMeasureHardnessCalculation(Entropy(reduction), score_threshold)


def BinaryEntropyCalculation(
    reduction: Callable[[np.ndarray], np.ndarray] = np.sum, score_threshold: float = 0.5
) -> UncertaintyMeasureHardnessCalculation:
    """
    Convenience function for hardness calculation with Entropy uncertainty measure

    Args:
        reduction: A function to calculate the hardness for the image based on the array of hardnesses for
            individual detected bounding boxes.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Returns:
        Uncertainty measure hardness calculation

    """
    return UncertaintyMeasureHardnessCalculation(BinaryEntropy(reduction), score_threshold)


class MonteCarloHardnessCalculation(UnsupervisedHardnessCalculation):
    """
    Unsupervised hardness calculation using Monte Carlo integration with stochastic ground truth model.

    A targeted hardness calculation where Monte Carlo simulation is used with a stochastic model for ground truth to
    approximate the hardness. The stochastic model is required because in this setting the ground truth is _uncertain_.

    Args:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
        reduction: Reduction function to be applied to samples of hardness, i.e. np.mean will yield expectation.
            A combination of np.mean and np.std can be used to obtain a lower bound

    Attributes:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
        reduction: Reduction function to be applied to samples of hardness, i.e. np.mean will yield expectation.
            A combination of np.mean and np.std can be used to obtain a lower bound
    """

    def __init__(
        self,
        hardness_definition: HardnessDefinition,
        pseudo_ground_truth_model: GroundTruthSampler,
        n_samples: int,
        score_threshold: float = 0.5,
        reduction: Callable[[np.ndarray], float] = np.mean,
    ):
        self.n_samples: int = n_samples
        self.pseudo_ground_truth_model: GroundTruthSampler = pseudo_ground_truth_model
        self.reduction: Callable[[np.ndarray], float] = reduction
        self.score_threshold: float = score_threshold
        self.hardness_definition: HardnessDefinition = hardness_definition

    def calculate_hardness(self, detection: DetectionType, image: torch.Tensor) -> float:
        """
        Estimate hardness for a single image

        Args:
            detection: Detections for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness
        """
        pseudo_ground_truth = self.pseudo_ground_truth_model.sample(detection, self.n_samples)

        filtered_detection = filter_pytorch_detections_by_score(detection, self.score_threshold)

        hardness_samples = [self.hardness_definition(filtered_detection, gt, image) for gt in pseudo_ground_truth]

        return self.reduction(np.array(hardness_samples))


class ScoreSamplingHardnessCalculation(MonteCarloHardnessCalculation):
    """
    Uses the score sampling pseudo ground truth model with Monte Carlo sampling to estimate hardness.

    Score sampling uses a Bernoulli random variable parameterised by bounding box scores to randomly discard boxes from
    the sampled pseudo ground truth. Runtime is usually approximately n_samples x hardness_definition time

    Args:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
        reduction: Reduction function to be applied to samples of hardness, i.e. np.mean will yield expectation.
            A combination of np.mean and np.std can be used to obtain a lower bound

    Attributes:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
        reduction: Reduction function to be applied to samples of hardness, i.e. np.mean will yield expectation.
            A combination of np.mean and np.std can be used to obtain a lower bound
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
    """

    def __init__(
        self,
        hardness_definition: HardnessDefinition,
        n_samples: int,
        score_threshold: float = 0.5,
        reduction: Callable[[np.ndarray], float] = np.mean,
    ):
        pseudo_ground_truth_model = ScoreSampling()
        super().__init__(hardness_definition, pseudo_ground_truth_model, n_samples, score_threshold, reduction)


class FastScoreSamplingHardnessCalculation(UnsupervisedHardnessCalculation):
    """
    A highly efficient alternative to ScoreSamplingHardness which avoids association of ground truth and detections.

    Can be targeted to some specific hardness definitions without actually calling the hardness definition function.

    False negatives and false positives are computed by assuming that the sampled pseudo ground truth will always
    associate to the detection from which it was generated, and hence the expensive association algorithm can be
    avoided.

    Args:
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        hardness_name: The name of the hardness definition to be targeted. Must be "fp" (false positives),
            "fn" (false negatives) or "false".
        weighting_fn: Weighting function to be used for boxes in evaluation when computing hardness.
            See hardest.pycocotools.weightings for more details.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
        reduction: Reduction function to be applied to samples of hardness, i.e. np.mean will yield expectation.
            A combination of np.mean and np.std can be used to obtain a lower bound

    Attributes:
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        hardness_name: The name of the hardness definition to be targeted. Must be "fp" (false positives),
            "fn" (false negatives) or "false".
        weighting_fn: Weighting function to be used for boxes in evaluation when computing hardness.
            See hardest.pycocotools.weightings for more details.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
        reduction: Reduction function to be applied to samples of hardness, i.e. np.mean will yield expectation.
            A combination of np.mean and np.std can be used to obtain a lower bound
    """

    def __init__(
        self,
        n_samples: int,
        hardness_name: Union[Literal["fp"], Literal["fn"], Literal["false"]],
        weighting_fn: Callable[[BoxesType, MasksType, torch.Tensor], torch.Tensor] = no_weighting,
        score_threshold: float = 0.5,
        reduction: Callable[[np.ndarray], float] = np.mean,
    ):
        super().__init__()
        self.n_samples: int = n_samples
        self.reduction: Callable[[np.ndarray], float] = reduction
        self.score_threshold: float = score_threshold
        self.weighting_fn: Callable[[BoxesType, MasksType, torch.Tensor], torch.Tensor] = weighting_fn
        self.hardness_name: Union[Literal["fp"], Literal["fn"], Literal["false"]] = hardness_name

    def calculate_hardness(self, detection: DetectionType, image: torch.Tensor) -> float:
        """
        Estimate hardness for a single image
        Args:
            detection: Detections for a single image in torchvision format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness

        """
        keep_distribution = Bernoulli(probs=detection["scores"])
        true = keep_distribution.sample((self.n_samples,)).bool()

        instances: BoxesType = detection

        detection_mask = detection["scores"] >= self.score_threshold

        masks: MasksType = {
            "tp": torch.logical_and(detection_mask, true).numpy(),
            "fp": torch.logical_and(detection_mask, ~true).numpy(),
            "fn": torch.logical_and(true, ~detection_mask).numpy(),
        }
        weight = torch.stack(
            [
                self.weighting_fn(
                    instances,
                    MasksType(
                        tp=masks["tp"][idx, ...],
                        fp=masks["fp"][idx, ...],
                        fn=masks["fn"][idx, ...],
                    ),
                    image,
                )
                for idx in range(self.n_samples)
            ],
            0,
        )

        assert (weight >= 0).all()

        fp = (weight * (detection["scores"] >= self.score_threshold) * (~true)).sum(axis=-1)
        fn = (weight * (detection["scores"] < self.score_threshold) * true).sum(axis=-1)

        if self.hardness_name == "fp":
            hardness_samples = fp
        elif self.hardness_name == "fn":
            hardness_samples = fn
        elif self.hardness_name == "false":
            hardness_samples = fp + fn
        else:
            raise NotImplementedError(f"Hardness name {self.hardness_name} is unknown, must be fp, fn or false.")

        assert len(hardness_samples) == self.n_samples

        return self.reduction(np.array(hardness_samples))


class IntegratedCommitteeHardnessCalculation(ABC):
    """
    Hardness calculation integrating over a committee (i.e. a set) of different detectors.

    The commitee could be identical detectors as in an ensemble, or otherwise. The committee will be treated as
    providing a discrete distribution over detections. This can be used to obtain a discrete distribution over hardness
    which can then be integrated to provide an expectated hardness.

    Args:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Attributes:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
    """

    def __init__(
        self,
        hardness_definition: HardnessDefinition,
        pseudo_ground_truth_model: DiscreteCommittee,
        score_threshold: float = 0.5,
    ):
        super().__init__()
        self.hardness_definition: HardnessDefinition = hardness_definition
        self.pseudo_ground_truth_model: DiscreteCommittee = pseudo_ground_truth_model
        self.score_threshold: float = score_threshold

    def calculate_hardness(
        self, detection_list_committee: List[DetectionType], detection: DetectionType, image: torch.Tensor
    ) -> float:
        """
        Estimate hardness for a single image

        Args:
            detection_list_committee: List of detections for each member of the committee for the image
            detection: Detections for a single image for detector for which hardness is being evaluated in torchvision
                format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness

        """
        pseudo_ground_truth, probability_mass = self.pseudo_ground_truth_model.probability_mass(
            detection_list_committee
        )

        detection = filter_pytorch_detections_by_score(detection, self.score_threshold)

        hardness_samples = [
            self.hardness_definition(detections=detection, annotations=gt, image=image) for gt in pseudo_ground_truth
        ]

        return sum(np.array(hardness_samples) * np.array(probability_mass))

    def eval_dataset(
        self,
        detection_list_committee: List[Iterable[DetectionType]],
        detections: Iterable[DetectionType],
        images: Iterable[torch.Tensor],
    ) -> List[float]:
        """
        Scores a dataset by hardness.

        Args:
            detection_list_committee: List of list of detections for committee, outer list over images in dataset, inner
                list over committee members.
            detections: List of detections in torchvision format for detector for which hardness is being evaluated
                (should be obtained by running the detector with a low score threshold as low score detections are
                required for the algorithm to work correctly).
            images: List of images in torchvision format (C, H, W) (can be a generator if dataset is large)

        Returns:
            Estimated hardness of dataset
        """
        detection_list_committee_per_image = list(zip(*detection_list_committee))

        with tqdm(detections, desc=f"Calculating hardness with {self}, fraction of dataset processed") as bar:
            return list(
                itertools.starmap(self.calculate_hardness, zip(detection_list_committee_per_image, bar, images))
            )


class SampledCommitteeHardnessCalculation(ABC):
    """
    Hardness calculation by score sampling algorithm on detections from a committee (i.e. a set) of different detectors

    A combination of IntegratedCommitteeHardness and ScoreSamplingHardness.

    Args:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).

    Attributes:
        hardness_definition: Definition of hardness to be computed (function of detections, targets and image)
        n_samples: Number of samples to make from pseudo ground truth model, which are used to compute hardness
            samples
        pseudo_ground_truth_model: Produces samples of pseudo ground truth based on detections.
        score_threshold: The detector score threshold (proposed detections with a score greater than this are
            are considered confirmed).
    """

    def __init__(
        self,
        hardness_definition: HardnessDefinition,
        n_samples: int,
        pseudo_ground_truth_model: ScoreSamplingUniformCommittee,  # TODO: eventually generalise this to non uniform
        score_threshold: float = 0.5,
    ):
        super().__init__()
        self.n_samples: int = n_samples
        self.hardness_definition: HardnessDefinition = hardness_definition
        self.pseudo_ground_truth_model: ScoreSamplingUniformCommittee = pseudo_ground_truth_model
        self.score_threshold: float = score_threshold

    def calculate_hardness(
        self, detection_list_committee: List[DetectionType], detection: DetectionType, image: torch.Tensor
    ) -> float:
        """
        Estimate hardness for a single image

        Args:
            detection_list_committee: List of detections for each member of the committee for the image
            detection: Detections for a single image for detector for which hardness is being evaluated in torchvision
                format
            image: Image for which hardness is being computed in torchvision format (C, H, W)

        Returns:
            Hardness

        """
        detection = filter_pytorch_detections_by_score(detection, self.score_threshold)

        pseudo_ground_truth = self.pseudo_ground_truth_model.sample(detection_list_committee, n_samples=self.n_samples)

        hardness_samples = [
            self.hardness_definition(detections=detection, annotations=gt, image=image) for gt in pseudo_ground_truth
        ]

        return np.mean(hardness_samples)

    def eval_dataset(
        self,
        detection_list_committee: List[Iterable[DetectionType]],
        detections: Iterable[DetectionType],
        images: Iterable[torch.Tensor],
    ) -> List[float]:
        """
        Scores a dataset by hardness.
        Args:
            detection_list_committee: List of list of detections for committee, outer list over images in dataset, inner
                list over committee members.
            detections: List of detections for detector for which hardness is being evaluated in torchvision format
                (should be obtained by running the detector with a low score threshold as low score detections are
                required for the algorithm to work correctly).
            images: List of images in torchvision format (C, H, W) (can be a generator if dataset is large)

        Returns:
            Estimated hardness of dataset
        """
        detection_list_committee_per_image = list(zip(*detection_list_committee))
        with tqdm(detections, desc=f"Calculating hardness with {self}, fraction of dataset processed") as bar:
            return list(
                itertools.starmap(self.calculate_hardness, zip(detection_list_committee_per_image, bar, images))
            )



import os
from typing import List

from setuptools import find_packages
from setuptools import setup

NAME = "hardest"
__version__ = None

repository_dir = os.path.dirname(__file__)


def filter_requirements_args(requirements: List[str]) -> List[str]:
    """Filter out arguments from a requirements list."""
    return [line for line in requirements if not line.strip().startswith("--") and len(line.strip()) > 0]


with open(os.path.join(repository_dir, "README.md")) as fh:
    README = fh.read()

with open(os.path.join(repository_dir, "requirements.txt")) as fh:
    requirements = filter_requirements_args(fh.readlines())

with open(os.path.join(repository_dir, "tests-requirements.txt")) as fh:
    tests_requirements = filter_requirements_args(fh.readlines())

setup(
    name=NAME,
    description="The HARDness ESTimation package: ranks an object detection dataset by the expected hardness of the "
    "images.",
    long_description_content_type="text/markdown",
    long_description=README,
    author="Jonathan Sadeghi",
    author_email="jonathan.sadeghi@five.ai",
    url="",
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Operating System :: POSIX :: Linux",
    ],
    include_package_data=True,
    install_requires=requirements,
    package_dir={"": "src"},
    packages=find_packages("src"),
    version="0.0.0",  # TOOD: do this the github ci way
    test_suite="tests",
    tests_require=tests_requirements,
    python_requires=">=3.8",
    extras_require={"test": tests_requirements},
)
